/* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 *  TCP helper functions.
 *
 *  Copyright (C) 2007 Gerd Hoffmann <kraxel@redhat.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <glib.h>

#include "spice-util.h"
#include "tcp.h"

int tcp_verbose;

/* ------------------------------------------------------------------ */

static char *strfamily(int family)
{
    switch (family) {
    case PF_INET6: return "ipv6";
    case PF_INET:  return "ipv4";
    case PF_UNIX:  return "unix";
    }
    return "????";
}

int tcp_connect(struct addrinfo *ai,
                char *addr, char *port,
                char *host, char *serv)
{
    struct addrinfo *res, *e;
    struct addrinfo *lres, ask;
    char uaddr[INET6_ADDRSTRLEN + 1];
    char uport[33];
    char uhost[INET6_ADDRSTRLEN + 1];
    char userv[33];
    int sock, rc, opt=1;

    /* lookup peer */
    ai->ai_flags = AI_CANONNAME;
    if (0 != (rc = getaddrinfo(host, serv, ai, &res))) {
        if (tcp_verbose)
            SPICE_DEBUG("getaddrinfo (peer): %s", gai_strerror(rc));
        return -1;
    }
    for (e = res; e != NULL; e = e->ai_next) {
        if (0 != getnameinfo((struct sockaddr*)e->ai_addr, e->ai_addrlen,
                             uhost, INET6_ADDRSTRLEN, userv, 32,
                             NI_NUMERICHOST | NI_NUMERICSERV)) {
            if (tcp_verbose)
                SPICE_DEBUG("getnameinfo (peer): oops");
            continue;
        }
        if (-1 == (sock = socket(e->ai_family, e->ai_socktype,
                                 e->ai_protocol))) {
            if (tcp_verbose)
                SPICE_DEBUG("socket (%s): %s",
                        strfamily(e->ai_family), strerror(errno));
            continue;
        }
        setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
        if (NULL != addr || NULL != port) {
            /* bind local port */
            memset(&ask, 0, sizeof(ask));
            ask.ai_flags    = AI_PASSIVE;
            ask.ai_family   = e->ai_family;
            ask.ai_socktype = e->ai_socktype;
            if (0 != (rc = getaddrinfo(addr, port, &ask, &lres))) {
                if (tcp_verbose)
                    SPICE_DEBUG("getaddrinfo (local): %s",
                            gai_strerror(rc));
                continue;
            }
            if (0 != getnameinfo((struct sockaddr*)lres->ai_addr,
                                 lres->ai_addrlen,
                                 uaddr, INET6_ADDRSTRLEN, uport, 32,
                                 NI_NUMERICHOST | NI_NUMERICSERV)) {
                if (tcp_verbose)
                    SPICE_DEBUG("getnameinfo (local): oops");
                continue;
            }
            if (-1 == bind(sock, lres->ai_addr, lres->ai_addrlen)) {
                if (tcp_verbose)
                    SPICE_DEBUG("%s [%s] %s bind: %s",
                            strfamily(lres->ai_family), uaddr, uport,
                            strerror(errno));
                continue;
            }
        }
        /* connect to peer */
        if (-1 == connect(sock, e->ai_addr, e->ai_addrlen)) {
            if (tcp_verbose)
                SPICE_DEBUG("%s %s [%s] %s connect: %s",
                        strfamily(e->ai_family), e->ai_canonname, uhost, userv,
                        strerror(errno));
            close(sock);
            continue;
        }
        if (tcp_verbose)
            SPICE_DEBUG("%s %s [%s] %s open",
                    strfamily(e->ai_family), e->ai_canonname, uhost, userv);
        return sock;
    }
    return -1;
}
